@@ -44,7 +44,7 @@ struct RRational {
 #endif
 
 #ifndef RRATIONAL
-# define RRATIONAL(obj) (R_CAST(RRational)(obj))
+# define RRATIONAL(obj) ((struct RRational *)(obj))
 #endif
 
 #ifndef RRATIONAL_SET_NUM
@@ -64,7 +64,7 @@ struct RComplex {
 #endif
 
 #ifndef RCOMPLEX
-# define RCOMPLEX(obj) (R_CAST(RComplex)(obj))
+# define RCOMPLEX(obj) ((struct RComplex *)(obj))
 #endif
 
 #ifndef RCOMPLEX_SET_REAL
@@ -97,7 +97,7 @@ static ID idPow, idPLUS, idMINUS, idSTAR, idDIV, idGE;
 static ID id_eqeq_p, id_idiv, id_negate, id_to_f, id_cmp, id_nan_p;
 static ID id_each, id_real_p, id_sum, id_population, id_closed, id_edge;
 
-static VALUE sym_left, sym_right;
+static VALUE sym_auto, sym_left, sym_right;
 
 static VALUE cHistogram;
 
@@ -2102,50 +2102,67 @@ histogram_edge_bin_index(VALUE edge, VALUE rb_x, int left_p)
 }
 
 static void
-histogram_weights_push_values(VALUE weights, VALUE edge, VALUE values, int left_p)
+histogram_weights_push_values(VALUE bin_weights, VALUE edge, VALUE values, VALUE weight_array, int left_p)
 {
+  const VALUE one = INT2FIX(1);
+  long bi, i, n, n_bins, weighted = 0;
   VALUE x, cur;
-  long i, n, bi;
+
+  assert(RB_TYPE_P(edge, T_ARRAY));
+  assert(RB_TYPE_P(values, T_ARRAY));
 
   n = RARRAY_LEN(values);
+  n_bins = RARRAY_LEN(edge) - 1;
+
+  if (! NIL_P(weight_array)) {
+    assert(RB_TYPE_P(weight_array, T_ARRAY));
+    assert(RARRAY_LEN(weight_array) == n);
+    weighted = 1;
+  }
+
   for (i = 0; i < n; ++i) {
     x = RARRAY_AREF(values, i);
 
-    bi = histogram_edge_bin_index(edge, x, left_p);
-
-    cur = rb_ary_entry(weights, bi);
-    if (NIL_P(cur)) {
-      cur = INT2FIX(1);
+    VALUE w;
+    if (weighted) {
+      w = RARRAY_AREF(weight_array, i);
+      if (RB_TYPE_P(w, T_COMPLEX)) {
+        VALUE imag = RCOMPLEX(w)->imag;
+        if (! RTEST(f_zero_p(imag))) {
+          goto type_error;
+        }
+      }
+      else if (rb_obj_is_kind_of(w, rb_cNumeric)) {
+        if (!RTEST(f_real_p(w))) {
+          goto type_error;
+        }
+      }
+      else {
+        goto type_error;
+      }
     }
     else {
-      cur = rb_funcall(cur, idPLUS, 1, INT2FIX(1));
+      w = one;
     }
 
-    rb_ary_store(weights, bi, cur);
-  }
-}
+    bi = histogram_edge_bin_index(edge, x, left_p);
 
-static int
-opt_closed_left_p(VALUE opts)
-{
-  int left_p = 1;
+    if (0 <= bi && bi < n_bins) {
+      cur = rb_ary_entry(bin_weights, bi);
+      if (NIL_P(cur)) {
+        cur = w;
+      }
+      else {
+        cur = rb_funcall(cur, idPLUS, 1, w);
+      }
 
-  if (!NIL_P(opts)) {
-    VALUE closed;
-#ifdef HAVE_RB_GET_KWARGS
-    ID kwargs = id_closed;
-    rb_get_kwargs(opts, &kwargs, 0, 1, &closed);
-#else
-    closed = rb_hash_lookup2(opts, ID2SYM(id_closed), sym_left);
-#endif
-    left_p = (closed != sym_right);
-    if (left_p && closed != sym_left) {
-      rb_raise(rb_eArgError, "invalid value for :closed keyword "
-               "(%"PRIsVALUE" for :left or :right)", closed);
+      rb_ary_store(bin_weights, bi, cur);
     }
   }
+  return;
 
-  return left_p;
+type_error:
+      rb_raise(rb_eTypeError, "weight array must have only real numbers");
 }
 
 static inline long
@@ -2269,10 +2286,47 @@ ary_histogram_calculate_edge(VALUE ary, const long nbins, const int left_p)
   return edge;
 }
 
+static VALUE
+check_histogram_weight_array(VALUE weight_array, const long ary_len)
+{
+  if (weight_array == Qundef) return Qnil;
+
+  weight_array = rb_check_convert_type(weight_array, T_ARRAY, "Array", "to_ary");
+  if (RARRAY_LEN(weight_array) != ary_len) {
+    rb_raise(rb_eArgError, "weight array must have the same number of items as the receiver array");
+  }
+
+  return weight_array;
+}
+
+static VALUE
+check_histogram_edges(VALUE edges)
+{
+  if (edges == Qundef) return Qnil;
+  return rb_check_convert_type(edges, T_ARRAY, "Array", "to_ary");
+}
+
+static int
+check_histogram_left_p(VALUE closed)
+{
+  int left_p = (closed != sym_right);
+  if (left_p && closed != Qundef && closed != sym_left) {
+    rb_raise(rb_eArgError, "invalid value for :closed keyword "
+        "(%"PRIsVALUE" for :left or :right)", closed);
+  }
+  return left_p;
+}
+
 /* call-seq:
- *    ary.histogram(nbins=:auto, closed: :left)
+ *    ary.histogram(nbins=:auto, weight: nil, closed: :left)
  *
  * @param [Integer] nbins  The approximate number of bins
+ * @params [Array<Numeric>] weights
+ *   An optional weight array, that has the same length as the receiver.
+ *   `weight[i]` means the weight value of the i-th element in the receiver.
+ * @params [Array<Numeric>] edges
+ *   An optional edge array, that specify the bin edges.
+ *   This array must be sorted.
  * @param [:left, :right] closed
  *   If :left (the default), the bin interval are left-closed.
  *   If :right, the bin interval are right-closed.
@@ -2282,30 +2336,52 @@ ary_histogram_calculate_edge(VALUE ary, const long nbins, const int left_p)
 static VALUE
 ary_histogram(int argc, VALUE *argv, VALUE ary)
 {
-  VALUE arg0, opts, edge, weights;
-  int left_p;
-  long nbins, nweights, i;
+  VALUE arg0, kwargs, bin_weights;
+  long nbins, n_bin_weights, i;
 
-  rb_scan_args(argc, argv, "01:", &arg0, &opts);
-  if (NIL_P(arg0)) {
+  VALUE weight_array = Qnil;
+  VALUE edges = Qnil;
+  int left_p = 1;
+
+  rb_scan_args(argc, argv, "01:", &arg0, &kwargs);
+  if (NIL_P(arg0) || arg0 == sym_auto) {
     nbins = sturges(RARRAY_LEN(ary));
   }
   else {
     nbins = NUM2LONG(arg0);
   }
-  left_p = opt_closed_left_p(opts);
 
-  edge = ary_histogram_calculate_edge(ary, nbins, left_p);
+  if (!NIL_P(kwargs)) {
+    enum { kw_weights, kw_edges, kw_closed };
+    static ID kwarg_keys[3];
+    VALUE kwarg_vals[3];
 
-  nweights = RARRAY_LEN(edge) - 1;
-  weights = rb_ary_new_capa(nweights);
-  for (i = 0; i < nweights; ++i) {
-    rb_ary_store(weights, i, INT2FIX(0));
+    if (!kwarg_keys[0]) {
+      kwarg_keys[kw_weights] = rb_intern("weights");
+      kwarg_keys[kw_edges]  = rb_intern("edges");
+      kwarg_keys[kw_closed] = rb_intern("closed");
+    }
+
+    rb_get_kwargs(kwargs, kwarg_keys, 0, 3, kwarg_vals);
+
+    weight_array = check_histogram_weight_array(kwarg_vals[kw_weights], RARRAY_LEN(ary));
+    edges = check_histogram_edges(kwarg_vals[kw_edges]);
+    left_p = check_histogram_left_p(kwarg_vals[kw_closed]);
   }
 
-  histogram_weights_push_values(weights, edge, ary, left_p);
+  if (NIL_P(edges)) {
+    edges = ary_histogram_calculate_edge(ary, nbins, left_p);
+  }
 
-  return rb_struct_new(cHistogram, edge, weights,
+  n_bin_weights = RARRAY_LEN(edges) - 1;
+  bin_weights = rb_ary_new_capa(n_bin_weights);
+  for (i = 0; i < n_bin_weights; ++i) {
+    rb_ary_store(bin_weights, i, INT2FIX(0));
+  }
+
+  histogram_weights_push_values(bin_weights, edges, ary, weight_array, left_p);
+
+  return rb_struct_new(cHistogram, edges, bin_weights,
                        left_p ? sym_left : sym_right,
                        Qfalse);
 }
@@ -2315,6 +2391,10 @@ Init_extension(void)
 {
   VALUE mEnumerableStatistics;
 
+#ifdef HAVE_RB_EXT_RACTOR_SAFE
+  rb_ext_ractor_safe(true);
+#endif
+
 #ifndef HAVE_ENUM_SUM
   rb_define_method(rb_mEnumerable, "sum", enum_sum, -1);
 #endif
@@ -2348,6 +2428,9 @@ Init_extension(void)
 
   rb_define_method(rb_cArray, "histogram", ary_histogram, -1);
 
+  void Init_array_extension(void);
+  Init_array_extension();
+
   idPLUS = '+';
   idMINUS = '-';
   idSTAR = '*';
@@ -2367,6 +2450,7 @@ Init_extension(void)
   id_closed = rb_intern("closed");
   id_edge = rb_intern("edge");
 
+  sym_auto = ID2SYM(rb_intern("auto"));
   sym_left = ID2SYM(rb_intern("left"));
   sym_right = ID2SYM(rb_intern("right"));
 }
